/*
	         Java RayTracer
	     By Dean Camera, 2008
	http://www.fourwalledcubicle.com
*/

public class Ray
{
	private Vector3D Origin;
	private Vector3D Direction;

	public Ray(Vector3D Origin, Vector3D Direction)
	{
		this.Direction = Direction.UnitVector();

		// Calculate starting position for the new ray - move it EPSILON away from the given origin
		this.Origin = new Vector3D(this.Direction).Multiply(Constants.EPSILON).Add(Origin);
	}

	public Vector3D GetOrigin()
	{
		return Origin;
	}

	public Vector3D GetDirection()
	{
		return Direction;
	}

	private ColorVector CombineColors(ColorVector ShadeColor, ColorVector ReflectionColor, ColorVector RefractionColor)
	{
		ColorVector FinalColor = new ColorVector(ShadeColor);
	
		FinalColor.Add(ColorVector.Multiply(ShadeColor, ReflectionColor));
		FinalColor.Add(RefractionColor);
	
		return FinalColor;
	}

	public ColorVector TraceRay(int RecursionsRem)
	{
		// Check if recursion limit exceeded
		if (RecursionsRem == 0)
			return ColorVector.BLACK;
	
		// Get the details on the object closest to the ray origin
		CollisionDetails ClosestObject = RayTraceObjects();

		// Check if an object intersected the ray
		if (ClosestObject != null)
		{
			ColorVector ShadeColor, ReflectColor, RefractColor;

			// Decrement the number of recurions allowed for sub rays from this ray
			RecursionsRem--;

			// Determine point and reflection colors at the point of intersection
			ShadeColor   = ShadePixel(ClosestObject);
			ReflectColor = TraceReflectionRay(ClosestObject, RecursionsRem);
			RefractColor = TraceRefractionRay(ClosestObject, RecursionsRem);

			// Return color composed of the hit object's color at the given point, plus reflected and refracted colors
			return CombineColors(ShadeColor, ReflectColor, RefractColor);
		}
		else
		{
			// No object intersected, return background color
			return Constants.SCENE_TO_RENDER.GetBackgroundColor();
		}
	}
	
	private ColorVector TraceReflectionRay(CollisionDetails OriginatingObject, int RecursionsRem)
	{
		Vector3D ReflectionRayDirection = LightPhysics.GetReflectionDirection(OriginatingObject.GetCollisionNormal(), Direction);
	
		// Ray trace the reflection ray and return the result
		ColorVector ReflectColor = new Ray(OriginatingObject.GetCollisionPoint(), ReflectionRayDirection).TraceRay(RecursionsRem);

		if (ReflectColor == Constants.SCENE_TO_RENDER.GetBackgroundColor())
			return ColorVector.BLACK;
		
		// Extract the object material's reflection constant
		float ReflectConst = OriginatingObject.GetCollisionObject().GetMaterial().GetReflective();

		return ColorVector.Multiply(ReflectColor, ReflectConst);
	}

	private ColorVector TraceRefractionRay(CollisionDetails OriginatingObject, int RecursionsRem)
	{
		Vector3D RefractionRayDirection = LightPhysics.GetRefractionDirection(OriginatingObject.GetCollisionNormal(), Direction,
		                                                                      OriginatingObject.GetCollisionObject().GetMaterial().GetRefractionIndex());
		
		if (RefractionRayDirection == null)
		  return ColorVector.BLACK;

		// Ray trace the refraction ray and return the result
		ColorVector RefractColor = new Ray(OriginatingObject.GetCollisionPoint(), RefractionRayDirection).TraceRay(RecursionsRem);

		// Extract the object material's refraction constant
		float RefractConst = OriginatingObject.GetCollisionObject().GetMaterial().GetTranslucency();

		return ColorVector.Multiply(RefractColor, RefractConst);
	}

	private CollisionDetails RayTraceObjects()
	{
		// Get the scene objects and lights
		ITraceableObject[] SceneObjects = Constants.SCENE_TO_RENDER.GetSceneObjects();

		ITraceableObject ClosestCollisionObject   = null;
		float            ClosestCollisionDistance = Float.MAX_VALUE;
	
		// Loop through each scene object, test for ray collision
		for (ITraceableObject CurrObject : SceneObjects)
		{
			float CollisionDistance = CurrObject.GetCollisionDistance(this);

			// Test if the current object collision ocurred closer to the screen than any previous collision
			if (CollisionDistance < ClosestCollisionDistance)
			{
				ClosestCollisionDistance = CollisionDistance;
				ClosestCollisionObject   = CurrObject;
			}
		}
		
		if (ClosestCollisionDistance != Float.MAX_VALUE)
			return ClosestCollisionObject.GetCollisionDetails(this, ClosestCollisionDistance);
		else
			return null;
	}

	private ColorVector ShadePixel(CollisionDetails StartingDetails)
	{
		ITraceableObject[] SceneObjects   = Constants.SCENE_TO_RENDER.GetSceneObjects();
		ILightingObject[] LightingObjects = Constants.SCENE_TO_RENDER.GetSceneLights();

		// Get the Opacity constant of the material
		float OpacityConst = (1 - StartingDetails.GetCollisionObject().GetMaterial().GetTranslucency());

		// Extract starting object point color
		ColorVector ObjectPointColor = ColorVector.Multiply(StartingDetails.GetCollisionColor(), OpacityConst);

		ColorVector FinalPointColor  = ColorVector.Multiply(ObjectPointColor, Constants.SCENE_TO_RENDER.GetAmbientLight());

		for (ILightingObject CurrLight : LightingObjects)
		{
			// Determine a unit vector indicating the direction from the collision object to the light source
			Vector3D LightToIntersectionPointDirection = Vector3D.Subtract(CurrLight.GetLocation(), StartingDetails.GetCollisionPoint());

			// Create a new ray from the collision object to the light source
			Ray ShadowRay = new Ray(StartingDetails.GetCollisionPoint(), LightToIntersectionPointDirection);

			// Determine the closest object obscuring the path from the object to the light source
			CollisionDetails ObjectObscuringPath = ShadowRay.RayTraceObjects();
			
			// Check if self was intersected - if so, don't calculate further
			if (ObjectObscuringPath == StartingDetails.GetCollisionObject())
			  continue;
			
			// Determine distance from the starting point to the light
			float PointToLightDistance = Vector3D.Subtract(StartingDetails.GetCollisionPoint(), CurrLight.GetLocation()).GetLengthSquared();

			// Check to see if the new ray did not hit another object in the path from the collision point to the light
			if ((ObjectObscuringPath == null) || (ObjectObscuringPath.GetCollisionDepth() > PointToLightDistance))
			{
				// No object hit - path is unobscured and the light may be applied
				FinalPointColor.Add(ColorVector.Multiply(ObjectPointColor, CurrLight.GetLightingDetails(StartingDetails, PointToLightDistance)));				
			}
		}
				
		return FinalPointColor;
	}
}
